/*
	$Id: task.h,v 3.2.2.2 2011/09/29 11:50:50 shelton Exp $
*/
#ifndef TASK_H
#define TASK_H

#include <setjmp.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>
#include <limits.h>

#include "list.h"
#include "message.h"
#include "collect.h"

#define NoWait      0L
#define WaitForever LONG_MAX

class Task;

class TaskList: public TList
{
protected:
  virtual void *getNext(void *item);
  virtual void *getPrev(void *item);
  virtual void setNext(void *item, void *next);
  virtual void setPrev(void *item, void *prev);
  virtual void freeItem(void *item);
public:
  TaskList(){;}
  Task *item();
  void makeZombie();
};

class TaskCollection: public SortedCollection
{
protected:
  virtual void freeItem( void *item );
  virtual int compare( void *key1, void *key2 );
  virtual void *keyOf( void *item );
public:
  TaskCollection();
  ~TaskCollection();
};

// Inter-task sync
class SemaBody;

class SemaList: public TList
{
protected:
  virtual void *getNext(void *item);
  virtual void *getPrev(void *item);
  virtual void setNext(void *item, void *next);
  virtual void setPrev(void *item, void *prev);
  virtual void freeItem(void *item);
public:
  SemaList();
  ~SemaList();
  SemaBody *item();
  void makeZombie();
};


class SemaBody
{
  friend class SemaList;
  friend class Sema;
  friend class Task;

  SemaBody *next;
  SemaBody *prev;
  int count;		// references counter
  int value;		//
  TaskList tasks;  	// blocked on semaphore tasks list by priorities
  void addSema(Sema *sem, Task *tp);
  void removeTask(Task *tp);

  SemaBody();
  ~SemaBody();
};



class Sema
{
friend class SemaBody;
friend class Task;

  int priority;
  SemaBody *sp;
  static int taskcount;

public:
  static int taskCount(){ return taskcount; }
  Sema(int prio, int closed = 0);  // priority to catch resource; 0 - minimal
  void setPrio(int newprio){ priority=newprio; }
  int  getPrio(){ return priority; }
  Sema( Sema &sema, int prio);
  Sema( Sema &sema );
  ~Sema();

  int close();	// wait and catch semaphore
  int open(); 	// unlock semaphore
  int waitClose(); // wait for locking semaphore by another somewho
};

/*
class SemaPipe
{
  Sema send;
  Sema recv;
public:
  SemaPipe(int pri=0): send(pri),recv(pri,1) {;}
  SemaPipe(SemaPipe&s): send(s.send), recv(s.recv) {;}
  SemaPipe(SemaPipe&s, int pri=0): send(s.send,pri), recv(s.recv,pri) {;}
  int send_beg() { return send.close(); }
  int send_end() { return recv.open();  }
  int recv_beg() { return recv.close(); }
  int recv_end() { return send.open();  }
};
*/

class Task
{
    friend class TaskHolder;
    friend class TaskCollection;
    friend class Message;
    friend class TaskMessage;
    friend class TaskList;
    friend class PrioList;
    friend class SemaBody;
    friend class SemaList;
    friend class Sema;
    friend class FileBuf;
    friend void  callTaskMain(Task *t);
#ifdef DJGPP
protected:
#endif
    const char  *tskname;       //  name of task

    // communication lists
    MessageList recvlist;	//  incoming messages (input queue)
    MessageList proclist;	//  processing messages

    int refcount;


    //  flags for task state
    unsigned  	 fInited:1      //  task was init'ed
		,fTimed:1       //  actived by timeout
		,fStarted:1     //  placed in ready-to-run list
		,isMapped:1     //  allocated stack in virtual memory
		,rfile:1
		,wfile:1
		,efile:1
		;
    enum {
	   Ready	//  task active
	  ,Zombie	//  task dead
	  ,Wait		//  wait file ready or timeout
	  ,Msg		//  wait message
	  ,Resp		//  wait answer
	  ,Sem		//  wait on semaphore
    } state:8;


    static jmp_buf  schEnv;     //  environment in scheduler
    jmp_buf     tskEnv;         //  environment for context switching
    int         stklen;         //  length of task's stack
    unsigned char *stack;       //  this task's stack space
    union{
    clock_t     wakeUp;         //  when to wake up if asleep
    int priority;		//  acted on semaphore queue
    };

    SemaBody *sem;
    Task *next;
    Task *prev;

    void  init();    //  set up stack and begin execution

    //  return to place of last yield()
    void  resume(int code = 1);

    static TaskCollection allTasks; // all tasks
    static TaskList readyTasks;
    static TaskList waitTasks;
    static TaskList msgTasks;    // waited for messages
    static TaskList respTasks;   // waited for answers on messages
    static SemaList semaList;    // semaphores list
    static TaskList zombieTasks;
    static Task * currTask;

    static int fullcount; // full number of tasks
    static u_long seqNo;
    u_long no;     // task number (unique)

    static fd_set readFiles;
    static fd_set writeFiles;
    static fd_set exceptFiles;

    fd_set rfileset;
    fd_set wfileset;
    fd_set efileset;

    void addTo(TaskList &list);
    void addToWait();  // with appropiate wakeup time place
    void addToReady();
    void addToZombie();
    void addToMsg();
    void addToResp();
    void removeFrom(TaskList &list);
    void RemoveFrom(TaskList &list); // do not check presence in list


    static Task *nextTask(); 	// set currTask

    static void testEvent()	{ waitEvent(0); }
    static void waitEvent(int block=1);

    virtual int canBeKilled();
    Task *parent;
    long chieldRes;

protected:
    Task(const char *tname, int stk = 1024);
    static int  yield0(); //  give up control & clear currTask
    static int _sendMessage(Task *receiver, Message &msg, int Block, int Free);

public:

    static void addref(Task *tp);
    static void delref(Task *tp);

    void start();
    static int run();
    static int shutDown();
    void suicide();
    virtual ~Task();
    virtual int identify(const char *data) 	{ return 1; }
    u_long getNo() {return no;}
    static Task* getCurrent(){ return currTask; }

    const char *name()          { return tskname; }

    static int  yield(); //  give up control

    int WakeUp(); // force stop of waiting

    // start chield and wait their finish
    long spawn(Task *chield);

    //  waited for ready-for-file or timeout exhausting
    //  !!! WAITING IS BECAUSE MESSAGE PROCESSED
    //    - arriving of message to task forced waiting (see handleMessage)

    static int  sleep(unsigned long msec);  // 1 - exhausting of timeout, 0 - force stop
    static void bsleep(unsigned long msec); // unaccepted waiting (do not processed incoming messages)
    static int  waitRead(int fd, unsigned long msec = WaitForever);
    static int  waitWrite(int fd, unsigned long msec = WaitForever);
    // the same as system "select" call
    static int  Select(int nfds,
		fd_set *readfds,
		fd_set *writefds,
		fd_set *exceptfds,
		struct timeval *timeout);

    // the main routine for each task
    virtual long main() = 0;
    // acted as destructor
    virtual void destroy();


    // take message from queue
    static Message *peekMessage(); // when no messages, return 0
    static Message *getMessage();  // when no messages, locked
    /*
      Asyncrohous messages processor
       - runs in task interrupt mode
         ( locked or gived control by yield() )
       - runs in called task stack
       - must not call lockable fuctions
       - common messages (with flag fCommon) passed only here
         because they shouldn't block at all
       - task messages passed here and next (when did not processed)
         - in incoming queue of task
       - here must processed events like "identifying", "redraw",
         "task information", etc. - all which connected with
         taking task state and do not connected with chaning
         task state
       - must return:
	 PROCESSED   when message processed
	 QUEUE	     when message must pass in incoming queue
	 UNBLOCK     when message must pass in incoming queue
		     and unlock task
	 BACK	     when messages must NOT pass in incoming queue
			(must call for unknown messages)
       - common messages (sent by sendAll ) must call only
	 BACK or PROCESSED, elsewhere messages will processed as BACK
    */
    enum MessPath {
	   PROCESSED,
	   QUEUE,
	   UNBLOCK,
	   BACK
	 };
    virtual MessPath handleMessage( Message *msg );

    /* message without answer waiting (destroy automatically after processing */
    static int sendMessage(u_long receiver, Message *msg);

    /* messages with answer waiting (automatically do not destroy) */
    static int sendMessage(u_long receiver, Message &msg);

    /* common message (automatically do not destroy */
    static int sendAll(Message &msg);

    static int sendMessage(Task *receiver, Message *msg);
    static int sendMessage(Task *receiver, Message &msg);


    /*
       sends reply to message ( resp destroy automatically)
       - when msg do not require answer, destroys
       !!! processing of any message must finished or by this call
       or by forward, else message sticks in processing list forever
    */
    static int respond(Message *msg, Message *resp=0);

    /* send message to another task */
    static int forward(Task  *receiver, Message *msg);
    static int forward(u_long receiver, Message *msg);

    static int kill(u_long task);

    static Task *findTask(u_long taskid);
    /* to adding messages from functions - system signals processors */
    void messToQueue( Message *msg );
};


// must called only from Task stack
extern "C" int  task_select(int nfds,
		fd_set *readfds,
		fd_set *writefds,
		fd_set *exceptfds,
		struct timeval *timeout);


class TaskHolder
{
   Task *tp;
public:
   void start(){ if (tp) tp->start(); }
   TaskHolder( Task *Tp );
   ~TaskHolder();

   Task *task(){ return tp; }
};


class TaskMessage: public Message
{
   Task *data;
 public:
   TaskMessage(Task *Data) { Task::addref(data=Data); 	}
   ~TaskMessage() 	{ if (data) Task::delref(data); }
   Task *item()		{ return data; }
   Task *getItem()   	{ Task *ret=data; data=0; return ret; }
   virtual long type();
};

class Timer : public Task
{
protected:
   unsigned closed:1;
   unsigned long msec;
public:
   Timer( Task *recv, long com, int sec=INT_MAX );
   long main();
   void setTime( int sec ) { msec = sec * 1000; }
   void open() { closed=0; }
   void close() { closed=1; }
   Task *reciever;
   long command;
};

#endif
